@@ -1,10 +1,14 @@
 from __future__ import annotations
+from io import BufferedReader, BytesIO
+import os
 
 import pickle
 import struct
 from dataclasses import dataclass, fields, is_dataclass
-from pathlib import Path
+from pathlib import Path, PosixPath, WindowsPath
 from typing import (
+    IO,
+    BinaryIO,
     Callable,
     Generic,
     List,
@@ -26,7 +30,11 @@ import pyarrow.parquet as pq
 from nptyping.ndarray import NDArrayMeta  # type: ignore
 from nptyping.shape_expression import get_dimensions  # type: ignore
 from ream.helper import has_dict_with_nonstr_keys
-from serde.helper import get_compression, get_open_fn
+from serde.helper import (
+    AVAILABLE_COMPRESSIONS,
+    get_filepath,
+    get_open_fn,
+)
 from tqdm import tqdm
 
 
@@ -179,6 +187,38 @@ class NumpyDataModel:
         setattr(newobj, field, value)
         return newobj
 
+    @classmethod
+    def concatenate(
+        cls,
+        models: list[NumpyDataModel],
+        merge_index: Optional[
+            dict[str, Callable[[list[NumpyDataModel]], dict | list | Index]]
+        ] = None,
+    ):
+        """Concatenate a list of NumpyDataModel objects into one object
+
+        Args:
+            models: a list of NumpyDataModel objects of the same class
+            merge_index: functions to merge an index, each key is the name of the index property. It's
+                optional if there is no index
+        """
+        metadata = cls._metadata
+        if metadata is None:
+            raise Exception(
+                f"{cls.__qualname__}.init() must be called before usage to finish setting up the schema"
+            )
+
+        merge_index = merge_index or {}
+
+        attrs = {}
+        for prop in metadata.array_props:
+            attrs[prop] = np.concatenate([getattr(m, prop) for m in models])
+
+        for prop in metadata.index_props:
+            attrs[prop] = merge_index[prop](models)
+
+        return cls(**attrs)
+
     def swap_(self, i: int, j: int):
         """Swap the position of two elements at index i and j.
         Note: This operator mutates the array
@@ -213,33 +253,38 @@ class NumpyDataModel:
     ):
         getattr(self, field)[i:j] = value
 
-    def save(self, file: Path, compression: Optional[str] = None):
+    def save(self, dir: Path, compression: Optional[AVAILABLE_COMPRESSIONS] = None):
+        """Save the data model to a directory containing 2 files: `data.parq` and `index.bin`.
+
+        Note: this function is carefully written so that the two files can be buffered to concatenate
+        multiple models to a single file using VirtualDir class. See `batch_save` for details.
+
+        Args:
+            dir: the directory to save the data model. Can be either Path or VirtualDir
+        """
         metadata = self._metadata
         if metadata is None:
             raise Exception(
                 f"{self.__class__.__qualname__}.init() must be called before usage to finish setting up the schema"
             )
 
-        compression = compression or get_compression(file)
         cols = {name: getattr(self, name) for name in metadata.array_props}
         for name in metadata.array2d_props:
             array2d = cols[name]
-            cols[name] = array2d[:, 0]
-            for i in range(1, array2d.shape[1]):
+            cols.pop(name)
+            for i in range(array2d.shape[1]):
                 cols[f"{name}_{i}"] = array2d[:, i]
 
+        dir.mkdir(parents=True, exist_ok=True)
         pq.write_table(
             pa.table(cols),
-            str(file),
+            dir / "data.parq",
             compression=compression or "NONE",
         )
 
         if len(metadata.index_props) > 0:
-            index_file = file.parent / (
-                file.stem
-                + (f".index.{compression}" if compression is not None else ".index")
-            )
-            with get_open_fn(index_file)(str(index_file), "wb") as f:
+            index_filename = get_filepath("index.bin", compression)
+            with get_open_fn(index_filename)(dir / index_filename, "wb") as f:
                 f.write(struct.pack("<I", len(metadata.index_props)))
                 for i, name in enumerate(metadata.index_props):
                     if metadata.index_prop_idxs[i][1] is not None:
@@ -250,22 +295,34 @@ class NumpyDataModel:
                     f.write(serindex)
 
     @classmethod
-    def load(cls, file: Path, compression: Optional[str] = None):
+    def batch_save(
+        cls,
+        batch: Sequence[NumpyDataModel],
+        dir: Path,
+        compression: Optional[AVAILABLE_COMPRESSIONS],
+    ):
+        """Save a list of numpy data models into a directory. Different from the save method of numpy data model,
+        data of multiple models will be concatenated into a single file.
+        """
+        with BatchFileManager() as filemanager:
+            virdir = VirtualDir(dir, filemanager=filemanager, mode="write")
+            for npmodel in batch:
+                npmodel.save(virdir, compression)
+                filemanager.flush()
+
+    @classmethod
+    def load(cls, dir: Path, compression: Optional[AVAILABLE_COMPRESSIONS] = None):
         metadata = cls._metadata
         if metadata is None:
             raise Exception(
                 f"{cls.__qualname__}.init() must be called before usage to finish setting up the schema"
             )
 
-        compression = compression or get_compression(file)
-        index_file = file.parent / (
-            file.stem
-            + (f".index.{compression}" if compression is not None else ".index")
-        )
         indices = []
 
-        if index_file.exists():
-            with get_open_fn(index_file)(str(index_file), "rb") as f:
+        if len(metadata.index_props) > 0:
+            index_filename = get_filepath("index.bin", compression)
+            with get_open_fn(index_filename)(dir / index_filename, "rb") as f:
                 n_indices = struct.unpack("<I", f.read(4))[0]
                 for i in range(n_indices):
                     size = struct.unpack("<I", f.read(4))[0]
@@ -276,7 +333,7 @@ class NumpyDataModel:
                         index = metadata.default_deserdict(f.read(size))
                     indices.append(index)
 
-        tbl = pq.read_table(str(file))
+        tbl = pq.read_table(dir / "data.parq")
         kwargs = {}
         if len(metadata.array2d_props) == 0:
             for name in metadata.array_props:
@@ -291,7 +348,7 @@ class NumpyDataModel:
                         names.append(col)
                     else:
                         newcols.append(col)
-                names.sort()
+                names.sort(key=lambda x: int(x.replace(name + "_", "")))
                 array2d = np.stack([tbl.column(s).to_numpy() for s in names], axis=1)
                 kwargs[name] = array2d
                 columns = newcols
@@ -313,26 +370,121 @@ class NumpyDataModel:
 
         return cls(**kwargs)
 
+    @classmethod
+    def batch_load(
+        cls, dir: Path, compression: Optional[AVAILABLE_COMPRESSIONS] = None
+    ):
+        with BatchFileManager() as filemanager:
+            virdir = VirtualDir(dir, filemanager=filemanager, mode="read")
+            output = []
+            while True:
+                output.append(cls.load(virdir, compression))
+                filemanager.flush()
+                if filemanager.is_all_read_done():
+                    break
+            return output
+
 
 @dataclass
 class NumpyDataModelContainer:
-    def save(self, dir: Path):
+    def save(self, dir: Path, compression: Optional[AVAILABLE_COMPRESSIONS] = None):
+        index_props = []
         for field in fields(self):
-            file = dir / f"{field.name}.parq"
-            getattr(self, field.name).save(file)
+            obj = getattr(self, field.name)
+            if isinstance(obj, (NumpyDataModel, NumpyDataModelContainer)):
+                obj.save(dir / field.name, compression)
+            else:
+                index_props.append((field.name, obj))
+
+        if len(index_props) > 0:
+            index_filename = get_filepath("index.bin", compression)
+            with get_open_fn(index_filename)(dir / index_filename, "wb") as f:
+                f.write(struct.pack("<I", len(index_props)))
+                for i, (name, obj) in enumerate(index_props):
+                    if isinstance(obj, Index):
+                        serindex = obj.to_bytes()
+                    else:
+                        serindex = pickle.dumps(obj)
+
+                    sername = name.encode()
+                    f.write(struct.pack("<I", len(sername)))
+                    f.write(sername)
+                    f.write(struct.pack("<I", len(serindex)))
+                    f.write(serindex)
 
     @classmethod
-    def load(cls, dir: Path):
+    def batch_save(
+        cls,
+        containers: Sequence[C],
+        dir: Path,
+        compression: Optional[AVAILABLE_COMPRESSIONS] = None,
+    ):
+        if len(containers) == 0:
+            raise ValueError("containers must not be empty")
+
+        with BatchFileManager() as filemanager:
+            virdir = VirtualDir(dir, filemanager=filemanager, mode="write")
+            for container in containers:
+                container.save(virdir, compression)
+                filemanager.flush()
+
+    @classmethod
+    def load(cls, dir: Path, compression: Optional[AVAILABLE_COMPRESSIONS] = None):
         assert is_dataclass(cls)
         type_hints: dict[str, type] = get_type_hints(cls)
         kwargs = {}
+        index_props = []
         for field in fields(cls):
-            file = dir / f"{field.name}.parq"
             fieldtype = type_hints[field.name]
-            assert issubclass(fieldtype, NumpyDataModel)
-            kwargs[field.name] = fieldtype.load(file)
+            if (ori_type := get_origin(fieldtype)) is not None:
+                fieldtype = ori_type
 
-        return NumpyDataModelContainer(**kwargs)
+            if issubclass(fieldtype, (NumpyDataModel, NumpyDataModelContainer)):
+                kwargs[field.name] = fieldtype.load(dir / field.name, compression)
+            else:
+                index_props.append(field)
+
+        if len(index_props) > 0:
+            index_filename = get_filepath("index.bin", compression)
+            n_npmodel = len(kwargs)
+
+            with get_open_fn(index_filename)(dir / index_filename, "rb") as f:
+                n_indices = struct.unpack("<I", f.read(4))[0]
+                for i in range(n_indices):
+                    size = struct.unpack("<I", f.read(4))[0]
+                    name = f.read(size).decode()
+                    size = struct.unpack("<I", f.read(4))[0]
+                    obj = f.read(size)
+                    obj_type = type_hints[name]
+                    if (ori_type := get_origin(obj_type)) is not None:
+                        obj_type = ori_type
+
+                    if issubclass(obj_type, Index):
+                        obj = obj_type.from_bytes(obj)
+                    else:
+                        obj = pickle.loads(obj)
+                    kwargs[name] = obj
+
+            assert len(kwargs) == n_npmodel + len(index_props)
+
+        return cls(**kwargs)
+
+    @classmethod
+    def batch_load(
+        cls, dir: Path, compression: Optional[AVAILABLE_COMPRESSIONS] = None
+    ):
+        with BatchFileManager() as filemanager:
+            virdir = VirtualDir(dir, filemanager=filemanager, mode="read")
+            output = []
+            while True:
+                output.append(cls.load(virdir, compression))
+                filemanager.flush()
+                if filemanager.is_all_read_done():
+                    break
+            return output
+
+
+C = TypeVar("C", bound=NumpyDataModelContainer)
 
 
 class ContiguousIndexChecker:
@@ -354,6 +506,95 @@ class ContiguousIndexChecker:
         return self
 
 
+class BatchFileManager:
+    def __init__(self):
+        self.open_write_files: dict[str, BinaryIO] = {}
+        self.open_read_files: dict[str, BufferedReader] = {}
+        self.pend_read_files: dict[str, BinaryIO] = {}
+        self.pend_write_files: dict[str, tuple[BytesIO, pa.PythonFile]] = {}
+
+    def create(self, filepath: str):
+        if filepath in self.pend_write_files:
+            raise Exception(
+                "Cannot request the same file twice. This is to prevent data corruption. Finishing writting data to the file and flushing it before requesting it again."
+            )
+        buf = BytesIO()
+        self.pend_write_files[filepath] = (buf, pa.PythonFile(buf))
+        if filepath not in self.open_write_files:
+            self.open_write_files[filepath] = open(filepath, "wb")
+
+        return self.pend_write_files[filepath][1]
+
+    def read(self, filepath: str) -> BinaryIO:
+        if filepath in self.pend_read_files:
+            raise Exception(
+                "Cannot request the same file twice. This is to prevent data corruption. Finishing reading data to the file and flushing it before requesting it again."
+            )
+
+        if filepath not in self.open_read_files:
+            self.open_read_files[filepath] = open(filepath, "rb")
+
+        file = self.open_read_files[filepath]
+        size = struct.unpack("<I", file.read(4))[0]
+        buf = file.read(size)
+
+        self.pend_read_files[filepath] = BytesIO(buf)
+        return self.pend_read_files[filepath]
+
+    def __enter__(self):
+        assert len(self.open_write_files) == 0 and len(self.open_read_files) == 0
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.flush()
+        for file in self.open_write_files.values():
+            file.close()
+        for file in self.open_read_files.values():
+            file.close()
+        self.open_read_files.clear()
+        self.open_write_files.clear()
+
+    def flush(self):
+        for filepath, (buf, file) in self.pend_write_files.items():
+            file = self.open_write_files[filepath]
+            buf = buf.getbuffer()
+            file.write(struct.pack("<I", buf.nbytes))
+            file.write(buf)
+
+        self.pend_write_files.clear()
+        self.pend_read_files.clear()
+
+    def is_all_read_done(self) -> bool:
+        return all(len(file.peek(1)) == 0 for file in self.open_read_files.values())
+
+
+class VirtualDir(Path):
+    __slots__ = ("filemanager", "mode")
+
+    _flavour = getattr(WindowsPath if os.name == "nt" else PosixPath, "_flavour")
+    filemanager: BatchFileManager
+    mode: Literal["read", "write"]
+
+    def __new__(cls, *args, **kwargs):
+        object = Path.__new__(cls, *args, **kwargs)
+        object.filemanager = kwargs["filemanager"]
+        object.mode = kwargs["mode"]
+        return object
+
+    def __truediv__(self, key: str) -> Union[pa.PythonFile, VirtualDir]:
+        key = str(key)
+        if key.find(".") == -1:
+            return VirtualDir(
+                str(self), key, filemanager=self.filemanager, mode=self.mode
+            )
+
+        filepath = str(super().__truediv__(key))
+        if self.mode == "write":
+            return self.filemanager.create(filepath)
+        else:
+            return self.filemanager.read(filepath)
+
+
 class NumpyDataModelHelper:
     """Generally contains helpers for a numpy data model that has a single index, which is a dictionary.
     The dictionary can have multiple levels, each level store a tuple of (start, end, nested_index),
@@ -362,6 +603,19 @@ class NumpyDataModelHelper:
     However, some helper methods can be used for other cases as well.
     """
 
+    @classmethod
+    def create_simple_index(
+        cls, names: list[str], npmodels: list[NumpyDataModel]
+    ) -> dict[str, tuple[int, int]]:
+        """Create a simple index that map each name to the index range of the corresponding numpy data model"""
+        index = {}
+        offset = 0
+        for name, npmodel in zip(names, npmodels):
+            size = len(npmodel)
+            index[name] = (offset, offset + size)
+            offset += size
+        return index
+
     @classmethod
     def stack(cls, npmodels: list[NumpyDataModel], keys: Sequence[str | int]):
         """This method concatenates a list of numpy data models of the same type into a single numpy data model.
@@ -374,7 +628,7 @@ class NumpyDataModelHelper:
         assert len(set(npmodel.__class__ for npmodel in npmodels)) == 1
         assert len(metadata.index_props) == 1
 
-        arrays = {prop: [] for prop in metadata.array_props}
+        arrays: dict = {prop: [] for prop in metadata.array_props}
         index = {}
         counter = 0
         for npmodel, key in zip(npmodels, keys):
@@ -389,8 +643,8 @@ class NumpyDataModelHelper:
             counter += size
 
         for prop, arr in arrays.items():
-            arrays[prop] = np.concatenate(arr)
-        arrays[metadata.index_props[0]] = index
+            arrays[prop] = np.concatenate(arr)  # type: ignore
+        arrays[metadata.index_props[0]] = index  # type: ignore
         return npmodels[0].__class__(**arrays)
 
     @classmethod
@@ -617,3 +871,14 @@ class NumpyDataModelHelper:
             obj[key] = tuple(obj[key])
 
         return index
+
+
+# dir = VirtualDir("/tmp", filetrack=FileTrack())
+# print(dir.name2file, dir.filetrack)
+# dir / "test.h5"
+# print(dir.name2file, dir.filetrack)
+# subdir = dir / "abc"
+# print(dir.name2file, dir.filetrack)
+# print(type(subdir))
+# print(subdir.name2file, subdir.filetrack)
+# # print((VirtualDir("/tmp") / "test.h5").name2file)
