@@ -42,10 +42,10 @@ def orjson_dumps(obj, **kwargs):
 
 
 def _logger_formatter_colorful(record):
-    clsname = "{extra[cls]}." if "cls" in record["extra"] else ""
+    name = "<cyan>{extra[name]}</cyan>:" if "name" in record["extra"] else ""
     return (
-        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>%s{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>\n{exception}"
-        % clsname
+        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | %s<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>\n{exception}"
+        % name
     )
 
 
@@ -60,6 +60,9 @@ def _logger_formatter(record):
 def _orjson_default(obj):
     if isinstance(obj, Path):
         return str(obj)
+    if isinstance(obj, set):
+        # so that the order is deterministic, and we can compare actor state by its serialized JSON
+        return sorted(obj)
     raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")
 
 
