@@ -1,4 +1,5 @@
 from ream.actor_state import ActorState
+from ream.actor_version import ActorVersion
 from ream.fs import FS, FSPath
 from ream.params_helper import (
     EnumParams,
@@ -10,13 +11,19 @@ from ream.actors.interface import Actor
 from ream.actors.base import BaseActor
 from ream.actor_graph import ActorGraph, ActorNode, ActorEdge
 from ream.dataset_helper import DatasetQuery, DatasetDict
-from ream.cache_helper import Cache
+from ream.cache_helper import Cache, CacheArgsHelper, Cacheable
 from ream.helper import configure_loguru
+from ream.data_model_helper import (
+    NumpyDataModel,
+    NumpyDataModelContainer,
+    NumpyDataModelHelper,
+)
 
 __all__ = [
     "Actor",
     "BaseActor",
     "ActorState",
+    "ActorVersion",
     "FS",
     "ReamWorkspace",
     "FSPath",
@@ -30,4 +37,9 @@ __all__ = [
     "DatasetDict",
     "configure_loguru",
     "Cache",
+    "CacheArgsHelper",
+    "Cacheable",
+    "NumpyDataModel",
+    "NumpyDataModelContainer",
+    "NumpyDataModelHelper",
 ]
